/* $Id: irc.c,v 1.2 1999/02/20 09:56:52 zap Exp $
 * -----------------------------------------------------
 * Copyright (c) 1998-99 Sebastian Kienzl <zap@riot.org>
 * -----------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * read 'COPYING' for details.
 */

#include "muh.h"
#include "irc.h"

/* #define DEBUG */

#define IRC_BUFFER 1024

struct hostent *hostinfo;
struct sockaddr_in addr;

#define TRACK 100
int track_socks[ TRACK ];

int highest_socket = 0;

int track_getfree()
{
    int i = 0;
    while( track_socks[ i ] && i < TRACK ) i++;
    return i;
}

int track_find( int no )
{
    int i = 0;
    while( track_socks[ i ] != no && i < TRACK ) i++;
    return i;
}

int track_highest()
{
    int h = 0;
    int i;
    for( i = 0; i < TRACK; i++ ) if( track_socks[ i ] > h ) h = track_socks[ i ];
    return h;
}

int sock_open()
/* returns -1 on error, otherwise socket-number */
{
    int i;
    i = socket( AF_INET, SOCK_STREAM, 0 );
    track_socks[ track_getfree() ] = i;
    highest_socket = track_highest();
    return i;
}

int sock_close( connection_type *connection )
/* closes socket */
{
    int i = 0;
    track_socks[ track_find( connection->socket ) ] = 0;
    highest_socket = track_highest();
    if( connection->socket ) i = close( connection->socket );
    connection->socket = 0;
    return i;
}

int rawsock_close( int socket )
{
    track_socks[ track_find( socket ) ] = 0;
    highest_socket = track_highest();
    return socket ? close( socket ) : 0;
}

int sock_setnonblock( int socket )
{
    return fcntl( socket, F_SETFL, O_NONBLOCK );
}

int sock_listen( int socket, int port, char *bindhost )
/* returns: bind() to port */
{
    int i = 1;
    setsockopt( socket, SOL_SOCKET, SO_REUSEADDR, &i, sizeof( i ) );
    sock_setnonblock( socket );

    bzero( &addr, sizeof( struct sockaddr_in ) );

    if( bindhost && ( ( hostinfo = gethostbyname( bindhost ) ) != NULL ) ) {
        memcpy( ( char * )&addr.sin_addr, hostinfo->h_addr, hostinfo->h_length );
        addr.sin_family = hostinfo->h_addrtype;
    }
    else {
        addr.sin_addr.s_addr = INADDR_ANY;
        addr.sin_family = AF_INET;
    }
    
    addr.sin_port = htons( ( u_short )port );
    i = bind( socket, ( struct sockaddr * )&addr, sizeof( struct sockaddr ) );
    return i | listen( socket, 1 );
}

int sock_accept( int socket, char *hostname )
{
    int temp;
    int store;

    temp = sizeof( struct sockaddr_in );
    store = accept( socket, ( struct sockaddr * )&addr, &temp );
    track_socks[ track_getfree() ] = store;
    highest_socket = track_highest();
    hostinfo = gethostbyaddr( ( char * )&addr.sin_addr.s_addr, sizeof( struct in_addr ), AF_INET );
    if( hostinfo ) strcpy( hostname, hostinfo->h_name );
    else strcpy( hostname, inet_ntoa( addr.sin_addr ) );
    return store;
}

int irc_write( connection_type *connection, char *format, ... )
{
    va_list va;
    char buffer[ IRC_BUFFER ];

    va_start( va, format );
    vsnprintf( buffer, IRC_BUFFER - 1, format, va );
    va_end( va );
    strcat( buffer, "\r\n" );
#ifdef DEBUG
    fprintf( stdout, ">>%03d>> %s", connection->socket, buffer );
#endif
    return write( connection->socket, buffer, strlen( buffer ) );
}

void irc_privmsg( connection_type *connection, char nickname[], char *format, ... )
{
    va_list va;
    char buffer[ IRC_BUFFER ];

    va_start( va, format );
    vsnprintf( buffer, IRC_BUFFER - 1, format, va );
    va_end( va );
    buffer[ BUFFERSIZE - 1 ] = 0;

    irc_write( connection, "PRIVMSG %s :%s", nickname, buffer );
}

void irc_notice( connection_type *connection, char nickname[], char *format, ... )
{
    va_list va;
    char buffer[ IRC_BUFFER ];

    va_start( va, format );
    vsnprintf( buffer, IRC_BUFFER, format, va );
    va_end( va );
    buffer[ BUFFERSIZE - 1 ] = 0;

    irc_write( connection, "NOTICE %s :%s", nickname, buffer );
}

int irc_read( connection_type *connection )
/* -1 -> fehler; 0 -> keine daten (falls non-blocking); 1 -> daten */
{
    int ret;
    connection->timer = 0;
    do {
	ret = read( connection->socket, connection->buffer + connection->offset, 1 );
	if( ret == 0 ) return -1; /* EOF */
	if( ret == -1 && errno == EAGAIN ) return 0;
        connection->offset++;
    } while( connection->buffer[ connection->offset - 1 ] != '\n' && connection->offset < BUFFERSIZE );
    connection->buffer[ connection->offset - 1 ] = 0; /* lf raus! */
    if( connection->buffer[ connection->offset - 2 ] == '\r' ) connection->buffer[ connection->offset - 2 ] = 0;
    /* bei servern immer, bei clients nicht */
#ifdef DEBUG
    fprintf( stdout, "<<%03d<< %s\n", connection->socket, connection->buffer );
#endif
    connection->offset = 0;
    return 1;
}

int irc_connect( connection_type *connection, server_type *server, irc_type *identity, char *bindto )
/* returns:
 0..........allright
 1..........kein socket konnte kreiert werden
 2..........server konnte nicht resolved werden
 3..........tcp/ip-connection zum server konnte nicht erffnet werden
 4..........write failed
 */
{
    int randport = 0, ri;
    
    connection->timer = 0;

    bzero( &addr, sizeof( struct sockaddr_in ) );
    if( ( connection->socket = sock_open() ) < 1 ) return 1;
    if( bindto && ( ( hostinfo = gethostbyname( bindto ) ) != NULL ) ) {
        memcpy( ( char * )&addr.sin_addr, hostinfo->h_addr, hostinfo->h_length );
    }
    else addr.sin_addr.s_addr = INADDR_ANY;

    do {
        for( ri = random() & 0xff; ri; ri-- )
            randport = ( random() & 0xffff ) | 1024;
        addr.sin_port = htons( randport );
    } while( bind( connection->socket, ( struct sockaddr * )&addr, sizeof( struct sockaddr ) ) );
    /* this weak random should do */

    if( ( hostinfo = gethostbyname( server->name ) ) == NULL ) return 2;
    memcpy( ( char * )&addr.sin_addr, hostinfo->h_addr, hostinfo->h_length );
    addr.sin_port = htons( ( u_short )server->port );
    addr.sin_family = hostinfo->h_addrtype;
    if( connect( connection->socket, ( struct sockaddr * )&addr, sizeof( struct sockaddr ) ) < 0 ) return 3;
    if( server->password ) irc_write( connection, "PASS %s", server->password );
    if( irc_write( connection, "NICK %s", identity->nickname ) < 0 ) return 4;
    if( irc_write( connection, "USER %s +i %s :%s", identity->username, server->name, identity->realname ) < 0 ) return 4;
    sock_setnonblock( connection->socket );
    return 0;
}

