/* $Id: irc.h,v 1.2 1999/02/20 09:56:52 zap Exp $
 * -----------------------------------------------------
 * Copyright (c) 1998-99 Sebastian Kienzl <zap@riot.org>
 * -----------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * read 'COPYING' for details.
 */

#ifndef ___IRC___
#define ___IRC___

#include "muh.h"

typedef struct {
    char *name;
    int	port;
    char password[ 15 ];
    int	working;
} server_type;
                
typedef struct {
    int	socket;
    int timer;
    char buffer[ BUFFERSIZE ];
    int	offset;
} connection_type;
                                
typedef struct {
    char nickname[ 50 ];
    char username[ 30 ];
    char realname[ 50 ];
} irc_type;

extern highest_socket;

int sock_open();
/* returns: socketnumber; -1 on error */

int sock_close( connection_type *connection );
/* returns: 0 on success; -1 on error */

int rawsock_close( int socket );
/* returns: 0 on success; -1 on error */

int sock_listen( int socket, int port, char *bindhost );
/* returns: 0 on success; -1 on error */

int sock_setnonblock( int socket );
/* returns: >= 0 on success; -1 on error */

int sock_accept( int socket, char *hostname );
/* returns: socketnumber; -1 on error */

int irc_write( connection_type *connection, char *format, ... );
/* returns: on success -> number of written bytes; -1 on error */

int irc_read( connection_type *connection );
/* returns: 1/0(no data (if blocking)) on success; -1 on error */

int irc_connect( connection_type *connection, server_type *server, irc_type *identity, char *bindto );
/* returns: 0 on success
 * errors: 1 - sock_open() failed
 *         2 - remotelookup failed
 *         3 - connect() failed
 *         4 - write() failed
 * IMPORTANT! connection->connected does NOT get set!
 */

void irc_privmsg( connection_type *connection, char nickname[], char *format, ... );
void irc_notice( connection_type *connection, char nickname[], char *format, ... );


/* v IRC-REPLY-CODES */

#define RPL_WELCOME 1
#define RPL_YOURHOST 2
#define RPL_SERVERIS 3
#define RPL_SERVERVER 4

#define RPL_MOTDSTART 375
#define RPL_MOTD 372
#define RPL_ENDOFMOTD 376

#define RPL_LUSERCLIENT 251
#define RPL_LUSEROP 252
#define RPL_LUSERUNKNOWN 253
#define RPL_LUSERCHANNELS 254
#define RPL_LUSERME 255

#define ERR_ERRONEUSNICKNAME 432
#define ERR_NICKNAMEINUSE 433
#define ERR_NICKUNAVAILABLE 437
#define ERR_NOPERMFORHOST 463
#define ERR_YOUREBANNEDCREEP 465

#define RPL_WHOISUSER 311
#define RPL_WHOISSERVER 312
#define RPL_WHOISOPERATOR 313
#define RPL_WHOISIDLE 317
#define RPL_WHOISCHANNELS 319
#define RPL_ENDOFWHOIS 318

#endif
