/* $Id: muh.c,v 1.10 1999/02/21 15:08:38 zap Exp $
 * -----------------------------------------------------
 * Copyright (c) 1998-99 Sebastian Kienzl <zap@riot.org>
 * -----------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * read 'COPYING' for details.
 */

/* #define DEBUG */

#include "muh.h"
#include "irc.h"
#include "parse.h"
#include "tools.h"
#include "messages.h"
#include "ignore.h"

FILE *messagelog = NULL;

server_type servers[ MAXSERVERS ];
irc_type identity;
status_type status;
cfg_type cfg = { 0, 1, 1, 1, 0, "", 0, "", "", "" };

typedef struct {
    int reply;
    int listen;
    int nickname;
    int antiidle;
} timer_type;
timer_type timers;

/*****verbindungs-bezogene variablen*************/
server_info i_server;
client_info i_client;

connection_type c_server;
connection_type c_client;
connection_type c_newclient;

int listensocket = 0;  	/* listensocket */

/*****variablen zum irc-daten interpretieren*****/
char *origin;
char *code;
char *param1;
char *param2;
char backup[ BUFFERSIZE ];


void escape()
{
    sock_close( &c_server );
    sock_close( &c_client );
    sock_close( &c_newclient );
    rawsock_close( listensocket );
    if( messagelog ) fclose( messagelog );
    unlink( FILE_PID );
    exit( 1 );
}

void sig_term()
{
    error( MUH_SIGTERM );
    escape();
}

void sig_empty() {}

void server_next()
{
    int i = i_server.current;
    
    sock_close( &c_server );
    c_server.socket = 0;
    
    do {
        i_server.current++;
        if( i_server.current >= i_server.amount ) i_server.current = 0;
    } while( !servers[ i_server.current ].working && ( i_server.current != i ) );

    if( !servers[ i_server.current ].working ) {
        error( MUH_OUTOFSERVERS );
        escape();
    }

    if( i == i_server.current ) {
        report( SOCK_RECONNECT, servers[ i_server.current ].name );
        sleep( 5 );
    }
}


void drop_client( connection_type *drop, char *reason )
{
    irc_write( drop, "ERROR :Closing Link: %s", reason );
    sock_close( drop );
}

void drop_all( char *reason )
{
    sock_close( &c_server );
    if( c_client.socket ) drop_client( &c_client, reason );
    if( c_newclient.socket ) drop_client( &c_newclient, reason );
}

int proceed_timer( int *timer, int warn, int exceed )
{
    (*timer)++;
    if( *timer == warn ) return 1;
    if( *timer >= exceed ) {
        *timer = 0;
        return 2;
    }
    return 0;
}

void client_left()
{
    if( cfg.leave ) report( MUH_LEAVING );
    status.suppress = 1;
    irc_write( &c_server, "WHOIS %s", status.nickname );
}


void check_timers()
{
    if( i_client.connected ) {
        switch( proceed_timer( &c_client.timer, 60, 120 ) ) {
        case 0:
            break;
        case 1:
            irc_write( &c_client, "PING :%s", i_server.realname );
            break;
        case 2:
            drop_client( &c_client, "Ping timeout" );
            error( CLNT_STONED );
            client_left();            
            i_client.connected = 0;
            break;
        }
    }
    else {
        if( cfg.antiidle && ( proceed_timer( &timers.antiidle, 0, 600 ) == 2 ) ) irc_write( &c_server, "PRIVMSG" );
    }

    if( i_server.connected == 2 ) {
        switch( proceed_timer( &c_server.timer, 60, 120 ) ) {
        case 0:
            break;
        case 1:
            irc_write( &c_server, "PING :%s", i_server.realname );
            break;
        case 2:
            drop_all( SERV_STONED );
            error( SERV_STONED );
            server_next();
            break;
        }
    }

    if( c_newclient.socket ) {
        switch( proceed_timer( &c_newclient.timer, 0, 20 ) ) {
        case 0:
        case 1:
            break;
        case 2:
            if( status.init ) {
                irc_write( &c_newclient, ":%s 464 %s :Password Incorrect", i_server.realname, identity.nickname );
                drop_client( &c_newclient, "Bad Password" );
            }
            else {
                drop_client( &c_newclient, "Ping timeout" );
            }
            error( CLNT_AUTHTO );
            break;
        }
    }

    if( proceed_timer( &timers.listen, 0, 10 ) == 2 ) status.allowconnect = 1;

    if( proceed_timer( &timers.reply, 0, 3 ) == 2 ) status.allowreply = 1;
    
    if( cfg.getnick && !status.got_nick && !i_client.connected && ( proceed_timer( &timers.nickname, 0, 3 ) == 2 ) )
        irc_write( &c_server, "NICK %s", identity.nickname );
}


int read_server()
{ /* diese prozedur is definitv zu lang, eventuell code-cleanup */
    int chop = 0, chvoice = 0;
    char onick[ 80 ];
    char ohostname[ 120 ];
    char *channel;
    int pass = 0;
    int rstate = irc_read( &c_server );

    if( rstate > 0 ) { /* new data...go for it! */
        rstate = 0;
        pass = 1;

        strncpy( backup, c_server.buffer, BUFFERSIZE );

        if( c_server.buffer[ 0 ] == ':' ) { /* reply */
            origin = c_server.buffer + 1;
            code = nextword( origin );
            param1 = nextword( code );
            param2 = nextword( param1 );

            param1[ pos( param1, ' ' ) ] = 0;
            code[ pos( code, ' ' ) ] = 0;
            origin[ pos( origin, ' ' ) ] = 0;
            /* hier wird der string in seine zeile 'zerlegt' (ned wirklich, nur gepointet) */
            
            switch( atoi( code ) ) {
            case RPL_WELCOME:
                /* reply 001 - server-welcome */
                i_server.connected++;
                strncpy( i_server.realname, origin, sizeof( i_server.realname ) );
                strncpy( status.nickname, param1, sizeof( status.nickname ) );
                strncpy( i_server.greeting[ 0 ], param2, sizeof( i_server.greeting[ 0 ] ) );
                sscanf( lastword( param2 ), "%*[^!]!%255s", ( char * )&status.idhostname );
                i_server.greeting[ 0 ][ lastpos( i_server.greeting[ 0 ], ' ' ) ] = 0;
                report( IRC_CONNECTED, i_server.realname );
                irc_write( &c_server, "MODE %s :+i", status.nickname );
                break;

            case RPL_YOURHOST:
            case RPL_SERVERIS:
            case RPL_SERVERVER:
                /* your-host, your-server-is, server-version replies */
                strncpy( i_server.greeting[ atoi( code ) - 1 ], param2, sizeof( i_server.greeting[ 0 ] ) );
                break;

            default:
                /* let's ignore other replies for now */
                break;
            }


            if( status.suppress ) {
                switch( atoi( code ) ) {
                case RPL_WHOISUSER:
                case RPL_WHOISSERVER:
                case RPL_WHOISOPERATOR:
                case RPL_WHOISIDLE:
                    pass = 0;
                    break;
                
                case RPL_WHOISCHANNELS:
                    pass = 0;
                    if( strncasecmp( status.nickname, param2, pos( param2, ' ' ) ) == 0 ) { /* whois-reply of our nick? */
                        backup[ 0 ] = 0;
                        while( ( param2 = nextword( param2 ) ) != 0 && strlen( param2 ) ) {
                            if( param2[ 0 ] == ':' ) param2++;
                            channel = ( char * )malloc( strlen( param2 ) + 1 );
                            strncpy( channel, param2, pos( param2, ' ' ) );
                            channel[ pos( param2, ' ' ) ] = 0;

                            if( ismodechan( channel ) ) {
                                if( channel[ 0 ] == '@' ) chop = 1; else chop = 0;
                                if( channel[ 0 ] == '+' ) chvoice = 1; else chvoice = 0;
                                strcpy( channel, channel + 1 );
                            }

                            if( !cfg.leave && i_client.connected ) {
                                irc_write( &c_client, ":%s!%s JOIN :%s", status.nickname, status.idhostname, channel );
                                if( chop ) irc_write( &c_client, ":%s!%s MODE %s +o %s", status.nickname, status.idhostname, channel, status.nickname );
                                if( chvoice ) irc_write( &c_client, ":%s!%s MODE %s +v %s", status.nickname, status.idhostname, channel, status.nickname );
                                if( channel[ 0 ] != '+' ) irc_write( &c_server, "TOPIC %s", channel );
                                irc_write( &c_server, "NAMES %s", channel );
                            }

                            strcat( backup, channel );
                            strcat( backup, "," );

                            free( channel );
                        };
                        backup[ strlen( backup ) - 1 ] = 0;
                        if( !i_client.connected ) {
                            if( cfg.leavemsg[ 0 ] ) irc_write( &c_server, "PRIVMSG %s :\1ACTION %s\1", backup, cfg.leavemsg );
                            if( cfg.leave ) irc_write( &c_server, "PART %s", backup );

                        }
                    }
                    break;

                case RPL_ENDOFWHOIS:
                    pass = 0;
                    status.suppress = 0;
                    break;
                }
            }

            if( strcmp( code, "NICK" ) == 0 ) { /* nickchange */
                if( ( strlen( status.nickname ) == pos( origin, '!' ) ) &&
                    ( strncasecmp( status.nickname, origin, pos( origin, '!' ) - 1 ) == 0 ) ) {
                    /* is that us who changed nick? */
                    strcpy( status.nickname, param1 + 1 );
                    if( strcasecmp( status.nickname, identity.nickname ) == 0 ) {
                        status.got_nick = 1;
                    }
                    else status.got_nick = 0;
                }
            }

            if( strcmp( code, "PONG" ) == 0 ) pass = 0;

            if( !i_client.connected ) { /* falls kein client da */
                switch( atoi( code ) ) {
                case ERR_ERRONEUSNICKNAME:
                    error( IRC_ERRIN, identity.nickname );
                    escape();
                    break;

                case ERR_NICKUNAVAILABLE:
                case ERR_NICKNAMEINUSE:
                    if( status.got_nick || ( i_server.connected != 2 ) ) {
                        randname( status.nickname, 8 );
                        report( IRC_NICKINUSE, identity.nickname, status.nickname );
                        irc_write( &c_server, "NICK %s", status.nickname );
                    }
                    status.got_nick = 0;
                    break;

                default:
                    break;
                }

                if( ( strcmp( code, "KICK" ) == 0 ) && ( strncasecmp( status.nickname, param2, pos( param2, ' ' ) - 1 ) == 0 ) &&
		    ( strlen( status.nickname ) == ( pos( param2, ' ' ) - 1 ) ) ) {
                    /* wtf?! we've been kicked of a channel! */
                    report( "%s kicked me out of %s (%s)!\n", origin, param1, nextword( param2 ) + 1 );
                }

                if( strcasecmp( param1, status.nickname ) == 0 ) { /* something has been sent to us */

                    if( ( strcmp( code, "PRIVMSG" ) == 0 ) || strcmp( code, "NOTICE" ) == 0 ) {
                        sscanf( origin, "%79[^!]!%*[^@]@%119s", ( char * )&onick, ( char * )&ohostname );
                        
                        if( param2[ 1 ] == '\1' ) { /* CTCP */
                            if( !is_ignore( ohostname, IGNORE_CTCP ) && status.allowreply ) {
                                upcase( param2 );
                                report( "received a CTCP %s from %s.\n", param2 + 1, origin );
                                if( strcmp( param2 + 2, "VERSION\1" ) == 0 )
                                    irc_notice( &c_server, onick, VERSIONREPLY );
                                if( strcmp( param2 + 2, "TIME\1" ) == 0 )
                                    irc_notice( &c_server, onick, TIMEREPLY );
                                if( strcmp( param2 + 2, "USERINFO\1" ) == 0 )
                                    irc_notice( &c_server, onick, USERINFOREPLY );
                                if( strncmp( param2 + 2, "PING", 4 ) == 0 )
                                    irc_notice( &c_server, onick, param2 + 1 );
                                if( strcmp( param2 + 2, "CLIENTINFO\1" ) == 0 )
                                    irc_notice( &c_server, onick, CLIENTINFOREPLY );
                                if( strcmp( param2 + 2, "FINGER\1" ) == 0 )
                                    irc_notice( &c_server, onick, FINGERREPLY );
                                add_ignore( ohostname, 6, IGNORE_CTCP );
                                status.allowreply = 0;
                            }
                        }
                        else { /* normale message/notice */
                            if( !is_ignore( ohostname, IGNORE_MESSAGE ) && status.allowreply ) {
                                if( cfg.logging && cfg.awaynotice[ 0 ] ) irc_notice( &c_server, onick, cfg.awaynotice );
                                add_ignore( ohostname, 120, IGNORE_MESSAGE );
                                status.allowreply = 0;
                            }
                            if( cfg.logging ) {
                                gettimestamp( backup );
                                fprintf( messagelog, "[%s](%s) %s\n", backup, origin, param2 + 1 );
                            }
                        }
                    }
                }
            } /* kein client da */
        }
        else { /* kommando */
            code = c_server.buffer;
            param1 = nextword( code ) + 1;
            code[ pos( code, ' ' ) ] = 0;
            
            if( strcmp( code, "PING" ) == 0 ) {
                irc_write( &c_server, "PONG :%s", param1 );
                pass = 0;
            }
            else
                if( strcmp( code, "ERROR" ) == 0 ) {
                    if( i_server.connected == 1 ) {
                        error( IRC_SERVERERROR, c_server.buffer + 7 );
                        servers[ i_server.current ].working = 0;
                    }
                    else error( IRC_SERVERERROR, c_server.buffer + 7 );
		    drop_all( SERV_ERR );                    
                }
        }
    }

    if( i_client.connected && pass ) irc_write( &c_client, "%s", backup );

    return rstate;
}


void fakeconnect()
{
    int i;
    FILE *check;
    
    irc_write( &c_client, ":%s 001 %s %s %s!~%s@%s", i_server.realname, status.nickname, i_server.greeting[ 0 ], status.nickname, i_client.username, i_client.hostname );
    for( i = 1; i < 4; i++ )
        irc_write( &c_client, ":%s %03d %s %s", i_server.realname, i + 1, status.nickname, i_server.greeting[ i ] );

    irc_write( &c_server, "LUSERS" );

    if( !cfg.leave ) report( MUH_REINTRODUCE );

    status.suppress = 1;
    irc_write( &c_server, "WHOIS %s", status.nickname );
    /* das kann man deswegen immer senden, da (falls muh brav aus allen channels gegangen ist) es eh keinen channel
     * reply im whois gibt (mach ich wegen rehashs - damit alles klappt falls cfg.leave umgestellt wird)
     */

    irc_write( &c_client, ":%s 375 %s :- muh -", i_server.realname, status.nickname );
    irc_write( &c_client, ":%s 372 %s :- running on server %s with nickname %s", i_server.realname, status.nickname, servers[ i_server.current ].name, status.nickname  );
    check = fopen( FILE_MESSAGES, "a" );
    if( ftell( check ) )
        irc_write( &c_client, ":%s 372 %s :- you have messages waiting. (read them with /muh read)", i_server.realname, status.nickname );
    else
        irc_write( &c_client, ":%s 372 %s :- you have no messages waiting.", i_server.realname, status.nickname );
    fclose( check );

    irc_write( &c_client, ":%s 376 %s :End of /MOTD command.", i_server.realname, status.nickname );
}

int read_newclient()
{
    int stat = irc_read( &c_newclient );
    
    if( stat > 0 ) {
        stat = 0;
        strncpy( code, c_newclient.buffer, 4 );
        code[ 4 ] = 0;
        upcase( code );

        if( strcmp( code, "PASS" ) == 0 ) {
            if( strncmp( c_newclient.buffer + 5, cfg.password, 50 ) == 0 ) {
                status.passok = 1;
            }
            status.init++;
        }
        if( strcmp( code, "NICK" ) == 0 ) status.init++;

        if( strcmp( code, "USER" ) == 0 ) status.init++;


        if( status.init >= 3 ) {
            if( !status.passok ) {
                if( i_client.connected ) irc_notice( &c_client, status.nickname, CLNT_AUTHFAILNOTICE, i_client.hostname );
                report( CLNT_AUTHFAIL );
            }
            else { /* client ist da! */
                param1 = nextword( c_newclient.buffer );
                if( param1 ) {
                    param1[ pos( param1, ' ' ) ] = 0;
                    strncpy( i_client.username, param1, sizeof( i_client.username ) );
                }
                else strcpy( i_client.username, "unknown" );

                report( CLNT_AUTHOK );

                if( i_client.connected ) {
                    drop_client( &c_client, "New connection established!" );
                    report( CLNT_DROP );
                }

                memcpy( &c_client, &c_newclient, sizeof( c_client ) );
                i_client.connected = 1;
                c_client.offset = 0;
                c_newclient.socket = 0;
                status.passok = 0;
                status.init = 0;
                fakeconnect();
            }
        }
    }
    return stat;
}

int read_client()
{
    int stat, corr;
    FILE *msglog;
    char logline[ 500 ];

    stat = irc_read( &c_client );
    if( stat > 0 ) { /* new data...go for it!! */
        stat = 0;
        strncpy( backup, c_client.buffer, BUFFERSIZE );
        code = c_client.buffer;
        param1 = nextword( code );
        if( param1 ) {
            param2 = nextword( param1 );
            code[ pos( code, ' ' ) ] = 0;
            if( param1[ 0 ] == ':' ) param1++;
        }
        else param1 = c_client.buffer;

        upcase( code );
        if( strcmp( code, "MUH" ) == 0 ) {
            upcase( param1 );
            corr = 0;
            if( strncmp( param1, "REA", 3 ) == 0 ) {
                corr++;
                msglog = fopen( FILE_MESSAGES, "a" );
                if( msglog && ftell( msglog ) ) {
                    freopen( FILE_MESSAGES, "r", msglog );
                    irc_notice( &c_client, status.nickname, "Playing messagelog..." );
                    while( fgets( logline, 499, msglog ) ) irc_notice( &c_client, status.nickname, logline );
                    irc_notice( &c_client, status.nickname, "End of messagelog." );
                    fclose( msglog );
                }
                else irc_notice( &c_client, status.nickname, CLNT_HAVENOMSGS );
            }
            if( strncmp( param1, "DEL", 3 ) == 0 ) {
                corr++;
                freopen( FILE_MESSAGES, "w", messagelog );
                irc_notice( &c_client, status.nickname, CLNT_KILLEDMSGS );
            }
            if( strncmp( param1, "REH", 3 ) == 0 ) {
                corr++;
                report( MUH_REHASH"\n" );
                irc_notice( &c_client, status.nickname, MUH_REHASH );
                strcpy( logline, identity.nickname );
                parse( MUHRC );
                if( strcmp( logline, identity.nickname ) ) status.got_nick = 0;
            }
            if( strncmp( param1, "JUM", 3 ) == 0 ) {
                corr++;
                drop_all( MUH_JUMP );
                report( MUH_JUMP"\n" );
                i_client.connected = 0;
                i_server.connected = 0;
                if( param2 ) {
                    new_server( param2 );
                    
                }
                server_next();
            }
            if( strncmp( param1, "DIE", 3 ) == 0 ) {
                corr++;
                drop_all( MUH_DIE_CL );
                report( MUH_DIE );
                escape();
            }
            if( !corr )
                irc_notice( &c_client, status.nickname, "Available commands: HELP READ DEL JUMP REHASH DIE" );
        }
        else
            if( strcmp( code, "QUIT" ) == 0 ) {
                drop_client( &c_client, "goodbye" );
                client_left();
                i_client.connected = 0;
                report( CLNT_LEFT );
            }
            else
                if( strcmp( code, "PONG" ) == 0 ) {
                    /* burp */
                }
                else
                    if( strcmp( code, "PING" ) == 0 ) {
                        irc_write( &c_client, ":%s PONG %s :%s", i_server.realname, i_server.realname, status.nickname );
                    }
                    else
                        irc_write( &c_server, "%s", backup );
    }
    return stat;
}

void run()
{
    fd_set rfds;
    struct timeval tv;

    while( 1 ) {
        if( !i_server.connected ) { /*schon wohin connected?*/
            report( SERV_TRYING, servers[ i_server.current ].name, servers[ i_server.current ].port );
            switch( irc_connect( &c_server, &servers[ i_server.current ], &identity, cfg.bind ) ) {
            case 0:
                report( SOCK_CONNECT, servers[ i_server.current ].name );
                i_server.connected = 1;
                break;
            case 1:
                error( SOCK_ERROPEN );
                escape();
                break;
            case 2:
                error( SOCK_ERRRESOLVE, servers[ i_server.current ].name );
                servers[ i_server.current ].working = 0;
                server_next();
                break;
            case 3:
                error( SOCK_ERRCONNECT, servers[ i_server.current ].name, strerror( errno ) );
                servers[ i_server.current ].working = 0;                
                server_next();
                break;
            case 4:
                error( SOCK_ERRWRITE, servers[ i_server.current ].name );
                server_next();
                break;
            default:
                break;
            }
        }

        tv.tv_usec = 0;
        tv.tv_sec = 1;

        FD_ZERO( &rfds );

        if( ( i_server.connected == 2 ) && status.allowconnect ) FD_SET( listensocket, &rfds );
        if( c_server.socket ) FD_SET( c_server.socket, &rfds );
        if( c_client.socket ) FD_SET( c_client.socket, &rfds );
        if( c_newclient.socket ) FD_SET( c_newclient.socket, &rfds );

        select( highest_socket + 1, &rfds, NULL, NULL, &tv );

        if( FD_ISSET( c_server.socket, &rfds ) ) { /* der server will was */
            if( read_server() < 0 ) {
                drop_all( SERV_DROPPED );
                error( SERV_DROPPED );
                i_client.connected = 0;
                i_server.connected = 0;
                server_next();
            }
        }
        else
            if( FD_ISSET( listensocket, &rfds ) ) { /* ein neuer client! */
                status.allowconnect = 0;
                c_newclient.socket = sock_accept( listensocket, i_client.hostname );
                c_newclient.offset = 0;
                if( !i_client.connected )
                    report( CLNT_CAUGHT, i_client.hostname );
                else {
                    report( CLNT_RECONNECT, i_client.hostname );
                }
            }
            else
                if( FD_ISSET( c_client.socket, &rfds ) ) { /* der client redet */
                    if( read_client() < 0 ) {
                        /* client hat connection gedroppt */
                        drop_client( &c_client, "!" );
                        client_left();
                        i_client.connected = 0;
                        report( CLNT_DROPPED );
                    }

                }
                else
                    if( FD_ISSET( c_newclient.socket, &rfds ) ) { /* der neue client sagt was */
                        if( read_newclient() < 0 ) {
                            /* newclient hat connection gedroppt */
                            drop_client( &c_newclient, "!" );
                            report( CLNT_DROPPEDUNAUTH );
                        }
                    }
                    else { /* this {} is entered *every* second */
                        check_timers();
                        process_ignores();
                    }

    }
}

void init()
{
    struct sigaction sv;

    sigemptyset( &sv.sa_mask );
    sv.sa_flags = 0;
    sv.sa_handler = sig_term;
    sigaction( SIGTERM, &sv, NULL );
    sigaction( SIGINT, &sv, NULL );
    sigaction( SIGKILL, &sv, NULL );
    sv.sa_handler = sig_empty;
    sigaction( SIGHUP, &sv, NULL );
    sigaction( SIGUSR1, &sv, NULL );
    sigaction( SIGPIPE, &sv, NULL );
    /* signalhandler hooken */
    
    umask( ~S_IRUSR & ~S_IWUSR );
    /* fuer logfile(s) */

    status.got_nick = 1; /* assume everythings fine */
    status.allowconnect = 1;
    
    srand( time( NULL ) );

    if( ( listensocket = sock_open() ) < 0 ) {
        error( SOCK_ERROPEN );
        escape();
    }
    if( sock_listen( listensocket, cfg.listenport, cfg.bind ) < 0 ) {
        error( SOCK_ERRBIND, cfg.listenport );
        escape();
    }
    report( SOCK_LISTENOK, cfg.listenport );
    if( cfg.bind[ 0 ] ) report( SOCK_BIND, cfg.bind );
    messagelog = fopen( FILE_MESSAGES, "a" );
#ifndef SETVBUF_REVERSED
    setvbuf( messagelog, NULL, _IONBF, 0 );
#else
    setvbuf( messagelog, _IONBF, NULL, 0 );
#endif
}

void main( int paramc, char *params[] )
{
    int  pid = 0;
    FILE *pidfile;
    char *home;
    
    fprintf( stdout, BANNER );

    if( ( home = getenv( "HOME" ) ) == 0 ) {
        error( MUH_NOHOME );
        escape();

    }

    cfg.home = ( char * )malloc( strlen( home ) + 1 );
    strcpy( cfg.home, home );
    
    realloc( cfg.home, strlen( cfg.home ) + strlen( WORKPATH ) + 1 );
    strcat( cfg.home, WORKPATH );
    chdir( cfg.home );
    
    report( MUH_PARSING );

    if( paramc >= 2 ) {
        if( !parse( params[ 1 ] ) ) escape();
    }
    else if( !parse( MUHRC ) ) escape();
    
    if( !check() ) escape();
    
    init();
#ifndef DEBUG
    pid = fork();
    if( pid < 0 ) {
        error( MUH_ERRFORK );
        escape();
    }
    if( pid == 0 ) {
        rename( FILE_LOG, FILE_OLDLOG );
        freopen( FILE_LOG, "w", stdout );
        freopen( FILE_LOG, "w", stderr );
        report( "mUH "VERSION" - starting log...\n" );
        report( SOCK_LISTENOK, cfg.listenport );
        if( cfg.bind[ 0 ] ) report( SOCK_BIND, cfg.bind );
#ifndef SETVBUF_REVERSED
        setvbuf( stdout, NULL, _IONBF, 0 );
        setvbuf( stderr, NULL, _IONBF, 0 );
#else
	setvbuf( stdout, _IONBF, NULL, 0 );
	setvbuf( stderr, _IONBF, NULL, 0 );
#endif       
    }
    if( pid ) {
        pidfile = fopen( FILE_PID, "w" );
        fprintf( pidfile, "%d\n", pid );
        fclose( pidfile );
        report( MUH_FORKED, pid );
        exit( 0 );
    }
#endif
    run();
}
