/* $Id: parse.c,v 1.2 1999/02/20 09:56:52 zap Exp $
 * -----------------------------------------------------
 * Copyright (c) 1998-99 Sebastian Kienzl <zap@riot.org>
 * -----------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * read 'COPYING' for details.
 */

#include "muh.h"
#include "irc.h"
#include "parse.h"
#include "tools.h"
#include "messages.h"

extern irc_type identity;
extern server_type servers[];
extern server_info i_server;
extern cfg_type cfg;

int new_server( char server[] );

typedef struct {
    char name[ 15 ];
    void *pointer;
    char scan[ 10 ]; /* scan-string */
    int set;  /* has been set (set it to 1 @ init-time if it's not a required keyword!)*/
    int special; /* if this is set, the param is passed on to a function */
    int (*func)( char [] ); /* 0 = error; <>0 = success*/
} struct_parse;

#define CONFIG_VARS 13
struct_parse parsewords[ CONFIG_VARS ] = {
    { "nick", &identity.nickname,  "%49s", 0, 0, 0 },
    { "realname", &identity.realname, "%49[^\n]", 0, 0, 0 },
    { "username", &identity.username, "%30s", 0, 0, 0 },
    { "server", 0, "", 0, 1, &new_server },
    { "listenport", &cfg.listenport, "%d", 0, 0, 0 },
    { "password", &cfg.password, "%19s", 0, 0, 0 },
    { "logging", &cfg.logging, "%d", 1, 0, 0 },
    { "leave", &cfg.leave, "%d", 1, 0, 0 },
    { "leavemsg", &cfg.leavemsg, "%59[^\n]", 1, 0, 0 },
    { "getnick", &cfg.getnick, "%d", 1, 0, 0 },
    { "bind", &cfg.bind, "%99s", 1, 0, 0 },
    { "antiidle", &cfg.antiidle, "%d", 1, 0, 0 },
    { "away", &cfg.awaynotice, "%199[^\n]", 1, 0, 0 },
};


int new_server( char server[] )
{
    int c = 0;
    int port;
    char temp[ 150 ];

    if( !sscanf( server, "%*[^:]:%d", &port ) ) port = 6667;

    while( c < MAXSERVERS && servers[ c ].name ) {
        if( ( strcmp( servers[ c ].name, server ) == 0 ) && ( servers[ c ].port == port ) ) return 1;
        c++;
    }

    c = sscanf( server, "%149[^:]:%d:%14s", temp, &servers[ i_server.amount ].port, (char *)&servers[ i_server.amount ].password );
    if( c == 1 ) servers[ i_server.amount ].port = 6667;
    servers[ i_server.amount ].name = ( char * )malloc( strlen( temp ) + 1 );
    strcpy( servers[ i_server.amount ].name, temp );
    servers[ i_server.amount ].working = 1;
    i_server.amount++;
    return c;
}

int check()
{
    int i;
    int ret = 1;
    for( i = 0; i < CONFIG_VARS; i++ ) {
        if( !parsewords[ i ].set ) {
            error( PARSE_MK, parsewords[ i ].name );
            ret = 0;
        }
    }
    return ret;
}


int parse( char filename[] )
{
    FILE *file;
    char buff[ 200 ];
    char cmd[ 20 ];
    char parm[ 100 ];
    int line = 0;
    int i;

    if( ( file = fopen( filename, "r" ) ) == NULL ) {
        error( FNF, filename );/*, strerror( errno ) );*/
        return 0;
    }

    while( fgets( buff, 199, file ) ) {
        line++;
        cmd[ 0 ] = 0;
        sscanf( buff, "%1s", cmd );
        if( cmd[ 0 ] == '\r' || cmd[ 0 ] == '\n' ) cmd[ 0 ] = 0;
        if( strlen( cmd ) && ( cmd[ 0 ] != '#' ) ) {
            sscanf( buff, "%19s %99[^\n\r]", cmd, parm );
            if( !strlen( parm ) ) error( PARSE_SE, line );
            i = 0;
            do {
                if( strcmp( cmd, parsewords[ i ].name ) == 0 ) {
                    if( parsewords[ i ].special ) {
                        if( parsewords[ i ].func( parm ) ) parsewords[ i ].set = 1;
                    }
                    else {
                        if( sscanf( parm, parsewords[ i ].scan, parsewords[ i ].pointer ) ) parsewords[ i ].set = 1;
                    }
                    if( !parsewords[ i ].set ) error( PARSE_SE, line );
                    i = 1000;
                }
                i++;
            } while( i <= CONFIG_VARS );
            if( i == (CONFIG_VARS+1) ) error( PARSE_UK, line );
            parm[ 0 ] = 0;
        }
    };
    return 1;
}
